// This software is not affiliated with Valve Corporation
//   We are not affiliated, associated, authorized, endorsed by, or in any way officially 
//   connected with Valve Corporation, or any of its subsidiaries or its affiliates. 
// 
//   All trademarks are property of their respective owners

typedef struct mdl_mesh mdl_mesh_t;

// High level API
//=======================================================================================================================

// Load model from vfs
int mdl_from_find_files( const char *mdlname, mdl_mesh_t *ctx );
void mdl_free( mdl_mesh_t *m );

// Set ctx to blank / 0,0
void mdl_error( mdl_mesh_t *m );

// Implementation
//=======================================================================================================================

struct mdl_mesh
{
	u16  *indices;	
	u32 	num_indices;
	
	float *vertices;
	u32 	num_vertices;
	
	// Bounding box
	boxf bounds;
};

#ifdef VALVE_IMPLEMENTATION

// VVD
//=======================================================================================================================
//StudioMDL constants

#define MAX_NUM_LODS 8
#define MAX_NUM_BONES_PER_VERT 3

#pragma pack(push, 1)
typedef struct
{
	float	weight[MAX_NUM_BONES_PER_VERT];
	char 	bone[MAX_NUM_BONES_PER_VERT];
	char 	numbones;
} 
boneWeight_t;

typedef struct
{
	boneWeight_t boneweights;
	float 		pos[3];
	float 		norm[3];	
	float 		uv[2];
} 
mstudiovertex_t;

typedef struct
{
	int id;
	int version;
	int checksum;
	int numLods;
	int numLodVertexes[MAX_NUM_LODS];
	int numFixups;
	int fixupTableStart;
	int vertexDataStart;
	int tangentDataStart;
} 
vertexFileHeader_t;

#pragma pack(pop)

mstudiovertex_t *GetVertexData( vertexFileHeader_t *t ) 
{	
	return (mstudiovertex_t *) ( (char *)t + t->vertexDataStart );
}

// VTX
//=======================================================================================================================

#pragma pack(push, 1)

typedef struct
{
	// these index into the mesh's vert[origMeshVertID]'s bones
	unsigned char boneWeightIndex[3];
	unsigned char numBones;

	unsigned short origMeshVertID;

	// for sw skinned verts, these are indices into the global list of bones
	// for hw skinned verts, these are hardware bone indices
	char boneID[3];
} 
VTXVertex_t;

enum StripGroupFlags
{
	STRIPGROUP_IS_FLEXED = 0x01,
	STRIPGROUP_IS_HWSKINNED = 0x02,
	STRIPGROUP_IS_DELTA_FLEXED = 0x04,
	STRIPGROUP_SUPPRESS_HW_MORPH = 0x08,	// NOTE: This is a temporary flag used at run time.
};

enum StripHeaderFlags_t {
	STRIP_IS_TRILIST	= 0x01,
	STRIP_IS_TRISTRIP	= 0x02
};

// A strip is a piece of a stripgroup which is divided by bones 
typedef struct 
{
	//Indices array
	int numIndices;
	int indexOffset;

	//Vertices array
	int numVerts;
	int vertOffset;

	short numBones;

	unsigned char flags;

	int numBoneStateChanges;
	int boneStateChangeOffset;
} 
VTXStripHeader_t;
// Bone state change inline code ommited

// a locking group
// a single vertex buffer
// a single index buffer
typedef struct
{
	// These are the arrays of all verts and indices for this mesh.  strips index into this.
	int numVerts;
	int vertOffset;

	int numIndices;
	int indexOffset;

	int numStrips;
	int stripOffset;

	unsigned char flags;
} 
VTXStripGroupHeader_t;

VTXVertex_t *pVertexVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (VTXVertex_t *)(((char *)t) + t->vertOffset) + i;
}
unsigned short *pIndexVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (unsigned short *)(((char *)t) + t->indexOffset) + i;
}
VTXStripHeader_t *pStripVTX( VTXStripGroupHeader_t *t, int i ) 
{ 
	return (VTXStripHeader_t *)(((char *)t) + t->stripOffset) + i; 
}

typedef struct
{
	int numStripGroups;
	int stripGroupHeaderOffset;

	unsigned char flags;
} 
VTXMeshHeader_t;
VTXStripGroupHeader_t *pStripGroupVTX( VTXMeshHeader_t *t, int i ) 
{ 
	return (VTXStripGroupHeader_t *)(((char *)t) + t->stripGroupHeaderOffset) + i; 
}

typedef struct
{
	//Mesh array
	int numMeshes;
	int meshOffset;

	float switchPoint;
} 
VTXModelLODHeader_t;
VTXMeshHeader_t *pMeshVTX( VTXModelLODHeader_t *t, int i ) 
{ 
	return (VTXMeshHeader_t *)(((char *)t) + t->meshOffset) + i; 
}

// This maps one to one with models in the mdl file.
typedef struct
{
	//LOD mesh array
	int numLODs;   //This is also specified in FileHeader_t
	int lodOffset;
} 
VTXModelHeader_t;
VTXModelLODHeader_t *pLODVTX( VTXModelHeader_t *t, int i ) 
{ 
	return (VTXModelLODHeader_t *)(((char *)t) + t->lodOffset) + i; 
}

typedef struct
{
	//Model array
	int numModels;
	int modelOffset;
} 
VTXBodyPartHeader_t;
VTXModelHeader_t *pModelVTX( VTXBodyPartHeader_t *t, int i ) 
{ 
	return (VTXModelHeader_t *)(((char *)t) + t->modelOffset) + i;
}

typedef struct
{
	// file version as defined by OPTIMIZED_MODEL_FILE_VERSION (currently 7)
	int version;

	// hardware params that affect how the model is to be optimized.
	int vertCacheSize;
	unsigned short maxBonesPerStrip;
	unsigned short maxBonesPerTri;
	int maxBonesPerVert;

	// must match checkSum in the .mdl
	int checkSum;

	int numLODs; // Also specified in ModelHeader_t's and should match

	// Offset to materialReplacementList Array. one of these for each LOD, 8 in total
	int materialReplacementListOffset;

	//Defines the size and location of the body part array
	int numBodyParts;
	int bodyPartOffset;
} 
VTXFileHeader_t;
VTXBodyPartHeader_t *pBodyPartVTX( VTXFileHeader_t *t, int i ) 
{
	return (VTXBodyPartHeader_t *)(((char *)t) + t->bodyPartOffset) + i;
}

	/*
		             .VTX file structure
		=============================================

		FileHeader
		  L	BodyParts::
			  L	Models::
				  L	LODS::
					  L	Meshes::
						  L	StripGroups::
							  L	VerticesTable[StudioMDL.Vertex]
							  L	IndicesTable[UINT16]
							  |
							  L	Strips::
								  L	Vertices[UINT16]
								  L	Indices[UINT16]
	*/

#pragma pack(pop)

u32 vtx_count_indices( VTXFileHeader_t *t )
{
	u32 indices = 0;
	
	for ( int bodyID = 0; bodyID < t->numBodyParts; ++bodyID )
	{
		VTXBodyPartHeader_t* pVtxBodyPart = pBodyPartVTX( t, bodyID );
		for ( int modelID = 0; modelID < pVtxBodyPart->numModels; ++modelID )
		{
			VTXModelHeader_t* pVtxModel = pModelVTX( pVtxBodyPart, modelID );

			int nLod = 0;
			VTXModelLODHeader_t *pVtxLOD = pLODVTX( pVtxModel, nLod );

			for ( int nMesh = 0; nMesh < pVtxLOD->numMeshes; ++nMesh )
			{
				VTXMeshHeader_t* pVtxMesh = pMeshVTX( pVtxLOD, nMesh );

				for ( int nGroup = 0; nGroup < pVtxMesh->numStripGroups; ++nGroup )
				{
					VTXStripGroupHeader_t* pStripGroup = pStripGroupVTX( pVtxMesh, nGroup );

					for ( int nStrip = 0; nStrip < pStripGroup->numStrips; nStrip++ )
					{
						VTXStripHeader_t *pStrip = pStripVTX( pStripGroup, nStrip );

						if ( pStrip->flags & STRIP_IS_TRILIST )
						{
							indices += pStrip->numIndices;
						}
					}
				}
			}
		}
	}
	
	return indices;
}

// MDL
//=======================================================================================================================

#pragma pack(push, 1)

typedef struct
{
	// base of external vertex data stores
	void			*pVertexData;
	void			*pTangentData;
} 
mstudio_modelvertexdata_t;

typedef struct
{
	// indirection to this mesh's model's vertex data
	int					unused_modelvertexdata; // 64b - Moved to follow num_LOD_Vertexes. 

	// used for fixup calcs when culling top level lods
	// expected number of mesh verts at desired lod
	int					numLODVertexes[MAX_NUM_LODS];
	
	mstudio_modelvertexdata_t *_the_death_ptr;
} 
mstudio_meshvertexdata_t;

typedef struct mstudiomodel_t mstudiomodel_t;

typedef struct
{
	int					material;
	int					modelindex; 
	int					numvertices;		// number of unique vertices/normals/texcoords
	int					vertexoffset;		// vertex mstudiovertex_t
	int					numflexes;			// vertex animation
	int					flexindex;
	// special codes for material operations
	int					materialtype;
	int					materialparam;
	// a unique ordinal for this mesh
	int					meshid;
	float					center[3];
	mstudio_meshvertexdata_t vertexdata;

	int					unused[6]; // remove as appropriate
} 
mstudiomesh_t;

// studio models
struct mstudiomodel_t
{
	char					name[64];
	int					type;
	float					boundingradius;

	int					nummeshes;	
	int					meshindex;

	int					numvertices;		// number of unique vertices/normals/texcoords
	int					vertexindex;		// vertex Vector
	int					tangentsindex;		// tangents Vector

	int					numattachments;
	int					attachmentindex;

	int					numeyeballs;
	int					eyeballindex;

	mstudio_modelvertexdata_t vertexdata;

	int					unused[8];		// remove as appropriate
};
mstudiomesh_t *studiomodel_pMesh( mstudiomodel_t *t, int i ) 
{ 
	return (mstudiomesh_t *)(((char *)t) + t->meshindex) + i; 
};

typedef struct
{
	int					sznameindex;
	int					nummodels;
	int					base;
	int					modelindex; // index into models array
} mstudiobodyparts_t;

mstudiomodel_t *mstudiobodyparts_pModel( mstudiobodyparts_t *t, int i ) 
{ 
	return (mstudiomodel_t *)(((char *)t) + t->modelindex) + i;
};

typedef struct {
	int					id;
	int					version;
	int					checksum;		// this has to be the same in the phy and vtx files to load!
	char					name[64];
	int					length;
	float					eyeposition[3];	// ideal eye position
	float					illumposition[3];	// illumination center
	float					hull_min[3];		// ideal movement hull size
	float					hull_max[3];
	float					view_bbmin[3];		// clipping bounding box
	float					view_bbmax[3];
	int					flags;
	int					numbones;			// bones
	int					boneindex;
	int					numbonecontrollers;		// bone controllers
	int					bonecontrollerindex;
	int					numhitboxsets;
	int					hitboxsetindex;
	int					numlocalanim;			// animations/poses
	int					localanimindex;		// animation descriptions
	int					numlocalseq;				// sequences
	int					localseqindex;
	int 					activitylistversion;
	int 					eventsindexed;
	// raw textures
	int					numtextures;
	int					textureindex;
	// raw textures search paths
	int					numcdtextures;
	int					cdtextureindex;
	// replaceable textures tables
	int					numskinref;
	int					numskinfamilies;
	int					skinindex;
	int					numbodyparts;
	int					bodypartindex;
	// queryable attachable points
	int					numlocalattachments;
	int					localattachmentindex;
	// animation node to animation node transition graph
	int					numlocalnodes;
	int					localnodeindex;
	int					localnodenameindex;
	int					numflexdesc;
	int					flexdescindex;
	int					numflexcontrollers;
	int					flexcontrollerindex;
	int					numflexrules;
	int					flexruleindex;
	int					numikchains;
	int					ikchainindex;
	int					nummouths;
	int					mouthindex;
	int					numlocalposeparameters;
	int					localposeparamindex;
	int					surfacepropindex;
	// Key values
	int					keyvalueindex;
	int					keyvaluesize;
	int					numlocalikautoplaylocks;
	int					localikautoplaylockindex;
	// The collision model mass that jay wanted
	float					mass;
	int					contents;
	// external animations, models, etc.
	int					numincludemodels;
	int					includemodelindex;
	// for demand loaded animation blocks
	int					szanimblocknameindex;
	int					numanimblocks;
	int					animblockindex;
	int					bonetablebynameindex;
	char					constdirectionallightdot;
	char					rootLOD;
	char					numAllowedRootLODs;
	char					unused[1];
	int					unused4; // zero out if version < 47
	int					numflexcontrollerui;
	int					flexcontrolleruiindex;
	float					flVertAnimFixedPointScale;
	int					unused3[1];
	int					studiohdr2index;
	int					unused2[1];
} 
studiohdr_t;

mstudiobodyparts_t	*studiohdr_pBodypart( studiohdr_t *t, int i ) 
{ 
	return (mstudiobodyparts_t *)(((char *)t) + t->bodypartindex) + i; 
};

#pragma pack(pop)

void mdl_free( mdl_mesh_t *m )
{
	free( m->indices );
	free( m->vertices );
}

void mdl_error( mdl_mesh_t *m )
{
	m->num_indices = 0;
	m->num_vertices = 0;
	m->indices = NULL;
	m->vertices = NULL;
}

int mdl_from_find_files( const char *mdlname, mdl_mesh_t *ctx )
{
	// Read entire files into memory (inline functions map memory)
	// .DX90.VTX
	char path[1024];
	strcpy( path, mdlname );
	csr_stripext( path );
	strcat( path, ".dx90.vtx" );
	VTXFileHeader_t *pVtxHdr = (VTXFileHeader_t *)valve_fs_get( path );
	
	if( !pVtxHdr )
	{
		mdl_error( ctx );
		return 0;
	}
	
	// .VVD
	strcpy( path, mdlname );
	csr_stripext( path );
	strcat( path, ".vvd" );
	vertexFileHeader_t *pVvdHdr = (vertexFileHeader_t *)valve_fs_get( path );
	
	if( !pVvdHdr )
	{
		free( pVtxHdr );
		mdl_error( ctx );
		return 0;
	}
	
	// .MDL
	strcpy( path, mdlname );
	csr_stripext( path );
	strcat( path, ".mdl" );
	studiohdr_t *pMdl = (studiohdr_t *)valve_fs_get( path );
	
	if( !pMdl )
	{
		free( pVtxHdr );
		free( pVvdHdr );
		mdl_error( ctx );
		return 0;
	}
	
	v3_copy( pMdl->hull_min, ctx->bounds[0] );
	v3_copy( pMdl->hull_max, ctx->bounds[1] );
	
	ctx->num_indices = vtx_count_indices( pVtxHdr );

	// Allocate and read indices
	ctx->indices = (u16 *)csr_malloc( ctx->num_indices * sizeof( u16 ) );
	ctx->num_indices = 0;	
	
	for ( int bodyID = 0; bodyID < pMdl->numbodyparts; ++bodyID )
	{
		// Body parts
		VTXBodyPartHeader_t* pVtxBodyPart = pBodyPartVTX( pVtxHdr, bodyID );
		mstudiobodyparts_t *pBodyPart = studiohdr_pBodypart( pMdl, bodyID );
		
		for ( int modelID = 0; modelID < pBodyPart->nummodels; ++modelID )
		{	
			// Models
			VTXModelHeader_t* pVtxModel = pModelVTX( pVtxBodyPart, modelID );
			mstudiomodel_t *pStudioModel = mstudiobodyparts_pModel( pBodyPart, modelID );

			int nLod = 0;
			VTXModelLODHeader_t *pVtxLOD = pLODVTX( pVtxModel, nLod );

			for ( int nMesh = 0; nMesh < pStudioModel->nummeshes; ++nMesh )
			{
				// Meshes
				VTXMeshHeader_t* pVtxMesh = pMeshVTX( pVtxLOD, nMesh );
				mstudiomesh_t* pMesh = studiomodel_pMesh( pStudioModel, nMesh );

				for ( int nGroup = 0; nGroup < pVtxMesh->numStripGroups; ++nGroup )
				{
					// Groups
					VTXStripGroupHeader_t* pStripGroup = pStripGroupVTX( pVtxMesh, nGroup );

					for ( int nStrip = 0; nStrip < pStripGroup->numStrips; nStrip++ )
					{
						// Strips
						VTXStripHeader_t *pStrip = pStripVTX( pStripGroup, nStrip );

						if ( pStrip->flags & STRIP_IS_TRILIST )
						{
							// Indices
							for ( int i = 0; i < pStrip->numIndices; i ++ )
							{
								u16 i1 =  *pIndexVTX( pStripGroup, pStrip->indexOffset + i );								
								ctx->indices[ ctx->num_indices ++ ] = pVertexVTX( pStripGroup, i1 )->origMeshVertID + pMesh->vertexoffset;
							}
						}
					}
				}
			}
		}
	}
	
	mstudiovertex_t *vertexData = GetVertexData( pVvdHdr );
	
	// Allocate vertex blob (XYZ|NRM|UV)
	ctx->num_vertices = pVvdHdr->numLodVertexes[0];
	ctx->vertices = (float *)csr_malloc( ctx->num_vertices * 8 * sizeof( float ) );
	
	for( int i = 0; i < ctx->num_vertices; i ++ )
	{
		mstudiovertex_t *vert = vertexData + i;
		memcpy( ctx->vertices + i * 8, vert->pos, 8 * sizeof(float) );
	}
	
	free( pVtxHdr );
	free( pVvdHdr );
	free( pMdl );
	
	return 1;
}

#endif
